from .audio import AudioEngine
from ..library.musicfileman import MusicFiles
import threading
import time
from typing import Optional
from pathlib import Path 

class Manager:
    def __init__(self):

        """
        Connects the audio engine, file system and user interface, manages playlist logic 
        and background monitoring
        """
        self.audio_engine = AudioEngine()
        self.music_files = MusicFiles()
        self.current_index = None
        self.is_playing = False 
        self.autoplay_enabled = True
        self._thread_started = False

    def start_playlist(self, index) -> None:
        """Plays a song based on its index in current library"""
        self.audio_engine.Pause()
        song_path = self.music_files.get_full_path(index)
        self.current_index = index
        self.audio_engine.Play(str(song_path))
        self.is_playing = True
    

    def next_song(self) -> None:
        """It plays next audio by using updated index"""
        total_songs = len(self.music_files.all_files)
        if total_songs > 0 :
            new_index = ( self.current_index + 1 ) % total_songs
            self.start_playlist(new_index)

    def pause_song(self) -> None:
        """Pauses the song"""
        if self.audio_engine.device.running:

            self.audio_engine.Pause()
        else:
            self.audio_engine.Resume()


    def start_thread(self) -> None:
        """Starts the thread for monitring the playback"""
        if not self._thread_started:

            t_thread = threading.Thread(target = self.watch_out, daemon = True)
            t_thread.start()
            self._thread_started = True 

    def watch_out(self) -> None:
        """ lo0p to handle song ending and autoplay"""
        while True:
            if self.is_playing: 
                if not self.audio_engine.device.running and not self.audio_engine.is_paused:
                    self.is_playing = False
                    self.audio_engine.reset_state()
                    if self.autoplay_enabled:
                        self.next_song() 
            time.sleep(0.3)

    def get_progress_percentage(self) -> float:
        """Returns percetange value of how much song completed using get progress method"""
        return self.audio_engine.get_progress()

    def play_by_path(self, path):
        """Play a specific file path directly from the browser."""
        self.audio_engine.Play(str(path))
        self.is_playing = True
        self.current_index = self.music_files.get_index_by_path(path)
        
if __name__ == '__main__':
    object1 = Manager()
    object1.start_playlist(1)
    object1.start_thread()
    input("enter to exit!")
    object1.audio_engine.close()



import miniaudio 
import time
from typing import Tuple


class AudioEngine:

    """ 
    High perforance audio backend using miniaudio and system-time tracking,
    This class hanldes low-level playback, state management and time calculations.
    """
    def __init__(self):
        self.device = miniaudio.PlaybackDevice()
        self.stream = None
        self.total_frames = 0
        self.sample_rate = 0
        self.is_paused = False
        self.start_time = 0.0
        self.elapsed_before_pause = 0.0

    def Play(self, file_path : str) -> None:
        """Loads and starts playback for a specific file using its file path"""

        self.reset_state()

        try: 

            info = miniaudio.get_file_info(file_path)
            self.total_frames = info.num_frames
            self.sample_rate = info.sample_rate
            self.elapsed_before_pause = 0.0

            self.stream = miniaudio.stream_file(file_path)
            self.start_time = time.time()
            self.device.start(self.stream)

            self.is_paused = False

        except FileNotFoundError: 
            raise FileNotFoundError(f"Audio file not found")
        except Exception as e:
            raise RuntimeError(f"Could not play: {e}")
        
    def Pause(self) -> None:
        """Pauses Playback and saves the time passed"""
        if self.device.running and not self.is_paused:
            self.elapsed_before_pause += (time.time() - self.start_time)
            # self.pause_time = time.time()
            self.is_paused = True
            self.device.stop()

    def Resume(self) -> None:
        """ Resumes the playback from the point that it was paused"""
        if self.is_paused and self.stream:
            self.start_time = time.time()
            self.device.start(self.stream)
            self.is_paused = False
        
    def reset_state(self):
        if self.device.running:
            self.device.stop()
        self.start_time = 0
        self.elapsed_before_pause = 0
        self.total_frames = 0
        self.is_paused = False


    def close(self) -> None:
        """ Releases Playbackdevice"""
        self.device.stop()
        self.device.close()

    def get_progress(self) -> float:
        """Returns playback progress in percentage"""
        if self.total_frames == 0 or not self.device.running:
            return 0
            
        elapsed = self.elapsed_before_pause
        if not self.is_paused:
            elapsed += (time.time() - self.start_time)
            
        duration = self.total_frames / self.sample_rate
        
        percent = (elapsed / duration) * 100
        return min(max(percent, 0), 100)

    def get_time_strings(self):
        """Returns (elapsed_str, total_str) -- for UI"""

        if self.total_frames == 0 or self.sample_rate == 0:
            return "00:00", "00:00"

        total_seconds = int(self.total_frames / self.sample_rate) if self.sample_rate > 0 else 0
        
        if self.is_paused:
            curr = int(self.elapsed_before_pause)
        else:
            curr = int(self.elapsed_before_pause + (time.time() - self.start_time))

        if curr > total_seconds:
            curr = total_seconds

        def fmt(seconds):
            seconds = max(0, int(seconds))
            mins, secs = divmod(seconds, 60)
            return f"{mins:02}:{secs:02}"
            
        return fmt(curr), fmt(total_seconds)


if __name__ == '__main__':
    music = AudioEngine()
    music.Play('/home/vaibahv/Downloads/samajwadi.mp3')

    while True:
        cmd = input(">> ")
        if cmd == 'p':
            music.Pause()
            print("stopped for now!")
        elif cmd == 'q':
            music.close()
            break



from pathlib import Path
import glob
import json
from typing import List, Dict, Union, Optional, Generator 

class MusicFiles:

    """
    Handles files system, metadata caching and directory navigation,
    Optimized for low-latency browsing in large directories
    """

    def __init__(self):
        self.root_path = Path.cwd()
        self.cache_file = self.root_path / ".vibe_cache.json"
        self.all_files = []
        self.allowed = ('.mp3', '.wav', '.flac', '.ogg')
        self.load_cache()

    def scanfiles(self) -> list:
        """ Recursive generator that yeilds path of audio files """

        try: 
            for path in self.root_path.rglob("*"):
                if path.is_file() and path.suffix.lower() in self.allowed:
                    yield str(path.relative_to(self.root_path))
        except PermissionError:
            pass
    

    def get_full_path(self,index:int) -> Optional[Path]:
        """ Retrives absolute path object"""
        if 0 <= index < len(self.all_files):
            relative_path_str = self.all_files[index]
            return self.root_path / relative_path_str
        return None

    def save_cache(self) -> None:
        """ Saves current library data in a JSON file"""
        try: 
            with open(self.cache_file, 'w', encoding='utf-8') as f:
                json.dump(self.all_files, f, indent=4)
        except Exception as e:
            print(f"erro saving cache: {e}")


    def load_cache(self, force_rescan = False) -> None:
        """ Loads cached paths """
        if self.cache_file.exists() and not force_rescan: 
            try:
                with open(self.cache_file, 'r', encoding= 'utf-8') as f:
                    self.all_files= json.load(f)
                if not self.all_files:
                    self.load_cache(force_rescan = True)
            except (json.JSONDecodeError, IOError):
                self.load_cache(force_rescan=True)
        else:
            self.all_files = list(self.scanfiles())
            self.save_cache()


    def get_full_path(self,index) -> Optional[Path]:
        """Retrives absolute Path object of using index"""
        if 0 <= index < len(self.all_files):
            relative_path_str = self.all_files[index]
            return self.root_path / relative_path_str
        return None

    def get_view_for_path(self, target_path: Path) -> List[Dict[str, Union[str, Path, bool]]]:
        """Returns a list of folders and music files in the target directory."""
        folders = []
        files = []
        
        if target_path != self.root_path:
            folders.append({
                "name": "⬅ .. [Back]", 
                "path": target_path.parent, 
                "is_dir": True
            })

        try:
            for p in sorted(target_path.iterdir()):
                if p.name.startswith('.'): continue
                
                if p.is_dir():
                    folders.append({
                        "name": f"📁 {p.name}", 
                        "path": p, 
                        "is_dir": True
                    })
                elif p.suffix.lower() in self.allowed:
                    files.append({
                        "name": f"🎵 {p.name}", 
                        "path": p, 
                        "is_dir": False
                    })
        except PermissionError:
            pass
            
        return folders + files
    
    def get_index_by_path(self, path) -> Optional[int]:
        """Finds where a file sits in playlist(CWD)"""
        try:
            rel_path = str(path.resolve().relative_to(self.root_path))
            return self.all_files.index(rel_path)
        except (ValueError, RuntimeError):
            return None 


if __name__ == '__main__':
    mp3 = MusicFiles()
    # list1 = mp3.scanfiles()
    # list2 = mp3.music_file()
    songs = mp3.get_file()
    print(songs)




import textual
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical, Container
from textual.widgets import Footer, Header, Button
from textual.widgets import Static, ListItem, ListView, Label, ProgressBar 
from ..audio.manager import Manager
from textual.reactive import reactive
import random



class VeridisTune(App):

    current_dir_items = reactive([])

    BINDINGS = [
    ("d", "toggle_dark", "Toggle Dark Mode"),
    ("space", "pause_play", "Pause/Play"),
    ("n", "next_song", "Next Track"),
    ]

    CSS_PATH = "style.tcss"

    def on_mount(self) -> None:
        self.manager = Manager()
        self.manager.start_thread()
        self.navigate_to(self.manager.music_files.root_path)
        self.set_interval(0.5, self.update_progress)

    def navigate_to(self, path):
        self.current_dir_items = self.manager.music_files.get_view_for_path(path)
        try:
            sidebar = self.query_one("#sidebar")
            sidebar.border_title = f"{path.name if path.name else '/'}"
        except:
            pass 
            
    
    def watch_current_dir_items(self, items: list) -> None:
            """Reactive watcher that repopulates the ListView whenever items change."""
            list_view = self.query_one('#song-list', ListView)
            list_view.clear()
            for item in items:
                list_view.append(ListItem(Label(item["name"])))

    def on_list_view_selected(self, event: ListView.Selected) -> None:
            index = event.list_view.index
            selected_item = self.current_dir_items[index]

            if selected_item["is_dir"]:
                self.navigate_to(selected_item["path"])
            else:
                self.manager.play_by_path(selected_item["path"])
                self.query_one("#now-playing").update(f"🎶 {selected_item['path'].name}")

    def compose(self) -> ComposeResult:
        yield Header() 
        with Horizontal():
            with Vertical(id="sidebar"):
                yield Label("Your library", id='sidebar-title')
                yield ListView(id='song-list')       
        
            with Vertical(id="main-zone"):
                yield Visualizer(id="Visualizer")
                yield Label("No song playing", id="now-playing")
            with Container(id="footer-zone"):
                yield Label("00:00", id="time-elapsed") 
                yield ProgressBar(total=100, show_eta = False, show_bar=True, id="progress")
                yield Label("00:00", id="time-total")   
        yield Footer()
        

    def update_progress(self):
            if hasattr(self, 'manager'):
                if not self.manager.is_playing:
                    self.query_one("#progress", ProgressBar).progress = 0
                    self.query_one("#time-elapsed").update("00:00")
                    self.query_one("#time-total").update("00:00")
                    return 
                self.query_one("#progress", ProgressBar).progress = self.manager.get_progress_percentage()
                elapsed_str, total_str = self.manager.audio_engine.get_time_strings()
                self.query_one("#time-elapsed").update(elapsed_str)
                self.query_one("#time-total").update(total_str)

    def action_pause_play(self) -> None:
        self.manager.pause_song()
        status = "Paused" if not self.manager.audio_engine.device.running else "Playing"
        self.notify(f"Music {status}")
        
    def action_next_song(self)-> None:
        self.manager.next_song()

class Visualizer(Static):
    def on_mount(self):
        self.set_interval(0.08, self.update_bars)

    def update_bars(self):
        if hasattr(self.app, "manager") and self.app.manager.is_playing:
            bars = ["▂", "▃", "▄", "▅", "▆", "▇", "█"]
            viz_str = "".join(random.choice(bars) for _ in range(40))
            self.update(viz_str)

        else:
            self.update("_" * 50)

if __name__ == "__main__":

    app = AudioPlayer()
    app.run()


from .interface.tui import VeridisTune
import time 
def main():
    time.sleep(0.1)
    try:
        app = VeridisTune()
        app.run()
    except Exception as e:
        print(f"{e}", file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()
